﻿module hdk.wrlayers;
private import hdk.sdlbase;
private import hdk.sdlinput;
private import hdk.sdlimage;
private import hdk.sdlttf;
private import hdk.sdlmixer;
private import hdk.oglbase;
private import hdk.ogltexture;
private import hdk.oglvertexbuffer;
private import hdk.oglshader;
private import hdk.math;
private import std.file;
private import hdk.wr;
private import std.cstream;
private import hdk.zipex;

auto abstract class WrLayer1 : WrLayers { mixin SdlBase; }
auto abstract class WrLayer2 : WrLayer1 { mixin SdlInput; }
auto abstract class WrLayer3 : WrLayer2 { mixin OglBase; }
auto abstract class WrLayer4 : WrLayer3 { mixin OglTextureMap_; }
auto abstract class WrLayer5 : WrLayer4 { mixin OglVertexBufferMap_; }
auto abstract class WrLayer6 : WrLayer5 { mixin OglShaderMap_; }

auto class WrLayer7 : WrLayer6
{
	void update() { error(__LINE__, "呼ぶな"); }
	
	bit doEvents()
	{
		swapBuffers();
		clear();
		
		SDL_Event event;
		while (pollEvent(&event))
		{
			if (event.type == SDL_QUIT) return false;
		}
		
		return true;
	}
	
	void addFileTexture(dchar[] fileName)
	{
		if (hasTexture(fileName)) return;
		
		uint width, height;
		uint[] pixels;
		if (ZipArchiveEx.isZip(fileName))
		{
			pixels = BufferedZipFile.readTexture(fileName, width, height);
		}
		else
		{
			pixels = readImage(fileName, width, height);
		}
		addTexture(fileName, pixels, width, height);
	}
	
	void addTextTexture(dchar[] text)
	{
		if (hasTexture(text)) return;
		uint[] textPixels = createTextTexture512x32(text);
		addTexture(text, textPixels, 512, 32);
	}
	
	void setShaderVector(dchar[] key, char[] shaderKey, Vector a)
	{
		super.setShaderVector(key, shaderKey, a.x, a.y, a.z, a.w);
	}
	
	void setShaderMatrixArray(dchar[] key, char[] shaderKey, Matrix[] ma)
	{
		super.setShaderMatrixArray(key, shaderKey, cast(float*)ma.ptr, ma.length);
	}
	
	void clearResource()
	{
		clearTextureMap();
		clearVertexBufferMap();
		BufferedZipFile.clear();
	}
	
	void drawPcnt(
		dchar[] shader, dchar[] vertexBuffer,
		dchar[] texture0, dchar[] texture1, dchar[] texture2, dchar[] texture3
		)
	{
		if (texture0 && !hasTexture(texture0)) loadTexture(texture0);
		if (texture1 && !hasTexture(texture1)) loadTexture(texture1);
		if (texture2 && !hasTexture(texture2)) loadTexture(texture2);
		if (texture3 && !hasTexture(texture3)) loadTexture(texture3);
		if (!hasVertexBuffer(vertexBuffer)) loadVertexBuffer(vertexBuffer);
		
		attachTexture(texture0, texture1, texture2, texture3);
		attachVertexBuffer(vertexBuffer);
		attachShader(shader);
		super.drawArrays(getAttachingVertexBufferLength());
		detachShader();
		detachVertexBuffer();
		detachTexture();
	}
	
	private void loadTexture(dchar[] a)
	{
		isFile(a) ? addFileTexture(a) : addTextTexture(a);
	}
	
	private bit isFile(dchar[] fileName)
	{
		if (ZipArchiveEx.isZip(fileName)) return true;
		try
		{
			isfile(toUTF8(fileName));
			return true;
		}
		catch (Object e) return false;
	}
	
	void drawPcntmw(
		dchar[] shader, char[] matrixIndex, char[] weight, dchar[] vertexBuffer,
		dchar[] texture0, dchar[] texture1, dchar[] texture2, dchar[] texture3
		)
	{
		if (texture0 && !hasTexture(texture0)) loadTexture(texture0);
		if (texture1 && !hasTexture(texture1)) loadTexture(texture1);
		if (texture2 && !hasTexture(texture2)) loadTexture(texture2);
		if (texture3 && !hasTexture(texture3)) loadTexture(texture3);
		if (!hasVertexBuffer(vertexBuffer)) loadVertexBuffer(vertexBuffer);
		
		uint miLocation = getShaderAttributeLocation(shader, matrixIndex);
		uint wLocation = getShaderAttributeLocation(shader, weight);
		
		attachTexture(texture0, texture1, texture2, texture3);
		attachVertexBuffer(vertexBuffer, miLocation, wLocation);
		attachShader(shader);
		super.drawArrays(getAttachingVertexBufferLength());
		detachShader();
		detachVertexBuffer();
		detachTexture();
	}
	
	uint totalResourceSize() { return totalTextureSize + totalVertexBufferSize; }
	
	void loadVertexBuffer(dchar[] fileName)
	{
		if (hasVertexBuffer(fileName)) return;
		if (!ZipArchiveEx.isZip(fileName)) error(__LINE__, "vertexBuffer read error");
		
		uint sizeOfVertex, vertexLength;
		float[] vertexBuffer = BufferedZipFile.readVertexBuffer(
			fileName, sizeOfVertex, vertexLength);
		
		if (sizeOfVertex == (3 + 4 + 3 + 2 + 2 + 1)) // pcntmw
		{
			addPcntmwVertexBuffer(fileName, vertexBuffer, vertexLength);
		}
		else if (sizeOfVertex == (3 + 4 + 3 + 2)) // pcnt
		{
			addPcntVertexBuffer(fileName, vertexBuffer, vertexLength);
		}
		else error(__LINE__, format("vertex size=%d", sizeOfVertex));
	}
	
	
	
	
	void drawRectangle(Vector color)
	{
		if (!hasShader("defaultRectangleShader"))
		{
			addShader("defaultRectangleShader",
				defaultRectangleVertexShader, defaultRectangleFragmentShader);
		}
		if (!hasVertexBuffer("defaultRectangleVertex"))
		{
			addPcntVertexBuffer("defaultRectangleVertex",
				defaultRectangleVertex, defaultRectangleVertexLength);
		}
		setShaderVector("defaultRectangleShader", "color", color);
		drawPcnt("defaultRectangleShader", "defaultRectangleVertex", null, null, null, null);
	}
	
	private const char[] defaultRectangleVertexShader = "
	uniform vec4 color;
	void main()
	{
		gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
		gl_FrontColor = color;
	}
	";
	private const char[] defaultRectangleFragmentShader = "
	void main()
	{
		gl_FragColor = gl_Color;
	}
	";
	
	private const float[] defaultRectangleVertex = [
		-0.5, -0.5, 0,  -0.5, 0.5, 0,  0.5, 0.5, 0,   -0.5, -0.5, 0,  0.5, 0.5, 0,  0.5, -0.5, 0,
		1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,
		0, 0, 0,  0, 0, 0,  0, 0, 0,   0, 0, 0,  0, 0, 0,  0, 0, 0,
		0, 1,  0, 0,  1, 0,   0, 1,  1, 0,  1, 1
		];
	private const uint defaultRectangleVertexLength = 6;
	
	void drawTexture(dchar[] texture, Vector color)
	{
		if (!hasShader("defaultTextureShader"))
		{
			addShader("defaultTextureShader",
				defaultTextureVertexShader, defaultTextureFragmentShader);
		}
		if (!hasVertexBuffer("defaultRectangleVertex"))
		{
			addPcntVertexBuffer("defaultRectangleVertex",
				defaultRectangleVertex, defaultRectangleVertexLength);
		}
		addFileTexture(texture);
		setShaderInt("defaultTextureShader", "texture0", 0);
		setShaderVector("defaultTextureShader", "color", color);
		drawPcnt("defaultTextureShader", "defaultRectangleVertex", texture, null, null, null);
	}
	
	private const char[] defaultTextureVertexShader = "
	uniform vec4 color;
	void main()
	{
		gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
		gl_TexCoord[0] = gl_MultiTexCoord0;
		gl_FrontColor = color;
	}
	";
	private const char[] defaultTextureFragmentShader = "
	uniform sampler2D texture0;
	void main()
	{
		gl_FragColor = texture2D(texture0, gl_TexCoord[0].xy) * gl_Color;
	}
	";
	
	void drawText(dchar[] text, Vector color)
	{
		if (!hasShader("defaultTextureShader"))
		{
			addShader("defaultTextureShader",
				defaultTextureVertexShader, defaultTextureFragmentShader);
		}
		if (!hasVertexBuffer("defaultRectangleVertex"))
		{
			addPcntVertexBuffer("defaultRectangleVertex",
				defaultRectangleVertex, defaultRectangleVertexLength);
		}
		addTextTexture(text);
		setShaderInt("defaultTextureShader", "texture0", 0);
		setShaderVector("defaultTextureShader", "color", color);
		drawPcnt("defaultTextureShader", "defaultRectangleVertex", text, null, null, null);
	}
	
	void postTerminableCommand(TerminableCommand cmd) { assert(false); }
	
	uint[] readBackBuffer() { return super.readBackBuffer(width, height); }
}
